#include "normalinterface.h"
#include "Shell.h"

void backup_file1(struct shell_state *sh, char *remote_path);

void copy_file(struct shell_state *sh, char *remote_path1, char *remote_path2);
void backup_file(struct shell_state *sh, char *remote_path);
void unbackup_file(struct shell_state *sh, char *remote_path1);
void restore_file(struct shell_state *sh, char *remote_path);
int list_directory_deep(string *args, struct shell_state *sh, int level);

//------------------------------------------------------------------------------------------

int n_deletePlaceholders(string *args, struct shell_state *sh)
{
	bSE_deletePlaceholders(sh);
	return SHELL_CONTINUE;
}

//------------------------------------------------------------------------------------------

int n_deleteAll_bSE(string *args, struct shell_state *sh)
{
	bSE_deleteAll_bSE(sh, (char *) args[1].c_str());
	return SHELL_CONTINUE;
}

//------------------------------------------------------------------------------------------

int n_activateSonOf(string *args, struct shell_state *sh)
{
	bSE_activateSonOf(sh);
	return SHELL_CONTINUE;
}

//------------------------------------------------------------------------------------------

int n_pwd(string *args, struct shell_state *sh)
{
	//cout << sh->remote_path << endl;
	my_cout(sh->remote_path.c_str(), kForceOut);
	
	return SHELL_CONTINUE;
}

int n_lpwd(string *args, struct shell_state *sh)
{
	//cout << sh->local_path << endl;
	my_cout(sh->local_path.c_str(), kForceOut);
	
	return SHELL_CONTINUE;
}

int n_setafc(string *args, struct shell_state *sh)
{
	string	arg1 = args[1];
	
	if (args[1] == "x") // CHK 2007/08/13
		arg1 = "com.apple.afc2";
	
	if (arg1 == "" )
	{
		my_cout("setafc: please provide a service name ('x' for com.apple.afc2).", kForceOutErr);
	} 
	else 
	{
		if (sh->shell_mode == SHELL_NORMAL)
			AFCConnectionClose(sh->conn);
		
		// Start AFC service
		
		my_cout("AMDeviceStartService AFC: ", AMDeviceStartService(sh->dev,
				CFStringCreateWithCString(NULL, arg1.c_str(), kCFStringEncodingASCII),
				&(sh->afch), NULL));

		// Open an AFC Connection
		my_cout("AFCConnectionOpen: ", AFCConnectionOpen(sh->afch, 0, &(sh->conn)));
	}
		
	return SHELL_CONTINUE;
}

int n_cd(string *args, struct shell_state *sh)
{
	// check for argument
	if(args[1] == "")
	{
		sprintf(gMsgStr, "//ERROR// cd: Please specify a path.");
		my_cout(gMsgStr, kForceOutErr);
		return SHELL_CONTINUE;
		
	} else {
		// assume relative path
		string temp = "/";
		
		if(args[1].c_str()[0]=='/')
			args[1] = args[1].substr(1);
		else
			temp = sh->remote_path;
		
		processRelativePath( &temp, &args[1] );
		
		if (dirExists(sh->conn, (char *)temp.c_str()))
			sh->remote_path=temp; 
		else
		{
			sprintf(gMsgStr, "cd: No such file or directory '%s'", (char *) temp.c_str());
			my_cout(gMsgStr, kForceOutErr);
		}
	}
	return SHELL_CONTINUE;
}

int n_lcd(string *args, struct shell_state *sh)
{
	string temp = "/";
	
	// check for argument
	if(args[1] == "")
	{
		my_cout("lcd: Please specify a path.", kForceOutErr);
		return SHELL_CONTINUE;
	} else {
		
		if(args[1].c_str()[0]=='/')
			args[1] = args[1].substr(1);
		else
			temp = sh->local_path;
		
		processRelativePath( &temp, &args[1] );
	}

	DIR *f = opendir(temp.c_str());
	if (f == NULL) 
	{
		sprintf(gMsgStr, "lcd: local file '%s' does not exist or is not a directory.", (char *) temp.c_str());
		my_cout(gMsgStr, kForceOutErr);
	} 
	else 
	{
		closedir(f);
		sh->local_path=temp;
	}

	return SHELL_CONTINUE;
}

int n_ls(string *args, struct shell_state *sh)
{
	//Remote LS userspace function
	struct afc_directory *dir;
	char *dirent;
	
	string ls_dir = "/";
	
	if ( args[1] != "" )
	{
		if(args[1].c_str()[0]=='/')
			args[1] = args[1].substr(1);
		else
			ls_dir = sh->local_path;
		
		processRelativePath( &ls_dir, &args[1] );
	
	} else {
		
		// we want to ls .
		ls_dir = sh->remote_path;
	
	}
	
	// open the dir for reading
	if (AFCDirectoryOpen(sh->conn, (char *) ls_dir.c_str(), &dir)) 
	{
		sprintf(gMsgStr, "ls: Directory '%s' does not exist.", (char *) ls_dir.c_str());
		my_cout(gMsgStr, kForceOutErr);
		return SHELL_CONTINUE;
	}
	
	// CHK 2007/08/12 added marker strings for easier parsing
	my_cout("//BEGIN_LS//", kForceOut);
	
	// list the dir
	while (1) 
	{
		AFCDirectoryRead(0, dir, &dirent);
		if (!dirent) break;

		// CHK 2007/08/12 was disabled
		if( strcmp(dirent,".") != 0 && strcmp(dirent,"..") != 0 )  // filter . and ..
		{
		
		#if 1
		{
			// CHK 2007/08/13 need to indicate whether it's a dir or not
			// do it by appending a slash if it is!

			char dircs[4096];
			
			sprintf(dircs, "%s/%s", (char *) ls_dir.c_str(), dirent);
			
			if (dirExists(sh->conn, dircs))	
			{
				sprintf(gMsgStr, "%s/", dirent);
				my_cout(gMsgStr, kForceOut);
			}
			else							
			{
				my_cout(dirent, kForceOut);
			}
		}
		#endif
		
		}
	}

	// CHK 2007/08/12 added marker strings for easier parsing
	my_cout("//END_LS//", kForceOut);

	AFCDirectoryClose(0, dir);
	return SHELL_CONTINUE;
}

int n_lsx(string *args, struct shell_state *sh)
{
	// CHK 2008/03/02
	// as n_ls except appends file size
	
	//Remote LS userspace function
	struct afc_directory *dir;
	char *dirent;
	
	string ls_dir = "/";
	
	if ( args[1] != "" )
	{
		if(args[1].c_str()[0]=='/')
			args[1] = args[1].substr(1);
		else
			ls_dir = sh->local_path;
		
		processRelativePath( &ls_dir, &args[1] );
	
	} else {
		
		// we want to ls .
		ls_dir = sh->remote_path;
	
	}
	
	// open the dir for reading
	if (AFCDirectoryOpen(sh->conn, (char *) ls_dir.c_str(), &dir)) 
	{
		sprintf(gMsgStr, "ls: Directory '%s' does not exist.", (char *) ls_dir.c_str());
		my_cout(gMsgStr, kForceOutErr);
		return SHELL_CONTINUE;
	}
	
	// CHK 2007/08/12 added marker strings for easier parsing
	my_cout("//BEGIN_LS//", kForceOut);
	
	// list the dir
	while (1) {
		AFCDirectoryRead(0, dir, &dirent);
		if (!dirent) break;

		// CHK 2007/08/12 was disabled
		if( strcmp(dirent,".") != 0 && strcmp(dirent,"..") != 0 )  // filter . and ..
		{
		
		#if 1
		{
			// CHK 2007/08/13 need to indicate whether it's a dir or not
			// do it by appending a slash if it is!

			char			dircs[4096];
			unsigned int	size;
			
			sprintf(dircs, "%s/%s", (char *) ls_dir.c_str(), dirent);
			
			if (dirExists(sh->conn, dircs))	
			{
				sprintf(gMsgStr, "%s/", dirent);
				my_cout(gMsgStr, kForceOut);
			}
			else							
			{
				size = get_file_size(sh->conn, dircs);
				sprintf(dircs, "%s//FILE_SIZE//%ld//", dirent, (long) size);
				my_cout(dircs, kForceOut);
			}
		}
		#endif
		
		}
	}

	// CHK 2007/08/12 added marker strings for easier parsing
	my_cout("//END_LS//", kForceOut);

	AFCDirectoryClose(0, dir);
	return SHELL_CONTINUE;
}

int n_lsd(string *args, struct shell_state *sh)
{return list_directory_deep(args, sh, 0);}

int list_directory_deep(string *args, struct shell_state *sh, int level)
{
	// CHK 2008/03/02
	// as n_ls except lists directories recursively
	
	//Remote LS userspace function
	struct afc_directory *dir;
	char *dirent;
	
	string ls_dir = "/";
	
	if ( args[1] != "" )
	{
		if(args[1].c_str()[0]=='/')
			args[1] = args[1].substr(1);
		else
			ls_dir = sh->local_path;
		
		processRelativePath( &ls_dir, &args[1] );
	
	} else {
		
		// we want to ls .
		ls_dir = sh->remote_path;
	
	}
	
	// open the dir for reading
	if (AFCDirectoryOpen(sh->conn, (char *) ls_dir.c_str(), &dir)) 
	{
		sprintf(gMsgStr, "ls: Directory '%s' does not exist.", (char *) ls_dir.c_str());
		my_cout(gMsgStr, kForceOutErr);
		return SHELL_CONTINUE;
	}
	
	if (! level)
	{
		// CHK 2007/08/12 added marker strings for easier parsing
		my_cout("//BEGIN_LS//", kForceOut);
	}

	sprintf(gMsgStr, "//LEVEL:%02d//", level);
	my_cout(gMsgStr, kForceOut);
	
	// list the dir
	while (1) {
		AFCDirectoryRead(0, dir, &dirent);
		if (!dirent) break;

		// CHK 2007/08/12 was disabled
		if( strcmp(dirent,".") != 0 && strcmp(dirent,"..") != 0 )  // filter . and ..
		{
		
		#if 1
		{
			// CHK 2007/08/13 need to indicate whether it's a dir or not
			// do it by appending a slash if it is!

			char	dircs[4096];
			
			sprintf(dircs, "%s/%s", (char *) ls_dir.c_str(), dirent);
			if (dirExists(sh->conn, dircs))	
			{
				string subArgs[2];
				
				sprintf(gMsgStr, "%s/", dirent);
				my_cout(gMsgStr, kForceOut);
				
				subArgs[1] = dircs;
				list_directory_deep(subArgs, sh, level+1);
			}
			else							
			{
				//size = get_file_size(sh->conn, dircs);
				//sprintf(dircs, "%s//FILE_SIZE//%ld//", dirent, (long) size);
				//cout << dircs << endl;
				
				my_cout(dirent, kForceOut);
			}
		}
		#endif
		
		}
	}
	
	if (! level)
	{
		// CHK 2007/08/12 added marker strings for easier parsing
		my_cout("//END_LS//", kForceOut);
	}
	
	AFCDirectoryClose(0, dir);
	return SHELL_CONTINUE;
}

int n_mkdir(string *args, struct shell_state *sh)
{
	string	path = "/";
	int		ret;
	
	if( args[1] == "" )
	{
		// assume we want the size of the current remote dir
		path = sh->remote_path;
	} else {
		
		if(args[1].c_str()[0]=='/')
		{
			args[1] = args[1].substr(1);
			processRelativePath(&path, &args[1]);
			
		} else {
			path = sh->local_path;
			processRelativePath( &path, &args[1] );
		}
	}
	
	ret = AFCDirectoryCreate(sh->conn, (char *) path.c_str());

	if (! ret) 
	{
		bSE_pushManifestItem(sh, kbSE_manifest_added, path.c_str(), nil);
		
		sprintf(gMsgStr, "//ADDED//<%s>//", (char *) path.c_str());
		my_cout(gMsgStr, kForceOut);
	}
	
	my_cout("mkdir: AFCDirectoryCreate", ret);
	return SHELL_CONTINUE;
}

int n_rmdir(string *args, struct shell_state *sh)
{
	afc_error_t	retval;
	string		path;
	
	if (args[1] == "")
	{
		// assume we want the size of the current remote dir
		path = sh->remote_path;
		
	} 
	else if ( args[1].at(0) == '/' ) 
	{
		#if kUSE_bSE
			bSE_delete(sh, (char *) args[1].c_str());
			return SHELL_CONTINUE;
		#endif
		
		// assume an abs path
		path = args[1];
	} 
	else
	{
		// assume we want the size of something relative to remote_path
		path = sh->remote_path;
		processRelativePath(&path, &args[1]);
	}
	
	retval = AFCRemovePath(sh->conn, (char *)path.c_str() );
	switch(retval)
	{
		case 0:
		{
			bSE_pushManifestItem(sh, kbSE_manifest_deleted, path.c_str(), nil);
			
			sprintf(gMsgStr, "//DELETED//<%s>//", (char *) path.c_str());
			my_cout(gMsgStr, kForceOut);
		}
		break;
		
		case 1:
			my_cout("rmdir: Could not remove directory, make sure it is empty.", kForceOutErr);
		default:
			my_cout("rmdir: AFCRemovePath returned unknown error code: ", retval);
	}

	return SHELL_CONTINUE;
}

int n_activate(string *args, struct shell_state *sh)
{
	my_cout("Implement me!", kForceOutErr);
	return SHELL_CONTINUE;
}

int n_deactivate(string *args, struct shell_state *sh)
{
	my_cout("AMDeviceDeavtivate: ", AMDeviceDeactivate(sh->dev));
	my_cout("deactivate: device deactivated.");
	return SHELL_CONTINUE;
}

int n_deviceinfo(string *args, struct shell_state *sh)
{
	char *key = NULL, *val = NULL;
	afc_dictionary *devinfo;
	afc_error_t ret;
	
	my_cout("deviceinfo: ", kForceOut);
	
	ret = AFCDeviceInfoOpen(sh->conn, &devinfo);
	if (ret == 0) {
		AFCKeyValueRead(devinfo, &key, &val);
		while (key && val) 
		{
			//cout << key << ": " << val << endl;
			sprintf(gMsgStr,"%s: %s", key, val);
			my_cout(gMsgStr, kForceOut);
			
			AFCKeyValueRead(devinfo, &key, &val);
		}
		AFCKeyValueClose(devinfo);
	} else
	{
		my_cout("error ", ret);
	}
	
	return SHELL_CONTINUE;
}

int n_readvalue(string *args, struct shell_state *sh)
{
	if( args[1] == "" )
	{
		my_cout("readvalue: Please enter a value to be read", kForceOutErr);
		return SHELL_CONTINUE;
	}
	
	__CFString * result = AMDeviceCopyValue(sh->dev, 0, 
						CFStringCreateWithCString(NULL, args[1].c_str(), kCFStringEncodingASCII));
	if (result)
	{
		//Windows handling code goes here!
		CFShow(result);
		CFStringGetCString(result, gLASTVALUE, sizeof(gLASTVALUE), kCFStringEncodingASCII);
	}
	else
	{
		sprintf(gMsgStr, "readvalue: Error reading value '%s'", args[1].c_str());
		my_cout(gMsgStr, kForceOut);
	}
	
	return SHELL_CONTINUE;
}

int n_enterrecovery(string *args, struct shell_state *sh)
{
	my_cout("AMDeviceEnterRecovery: ", AMDeviceEnterRecovery(sh->dev));
	return SHELL_CONTINUE;
}

int n_reconnect(string *args, struct shell_state *sh)
{
	my_cout("shell: Terminating and *not* waiting for device reconnect.", kForceOut);
	return SHELL_TERMINATE; // CHK 2007/08/08 was SHELL_WAIT;
}

int n_startservice(string *args, struct shell_state *sh)
{
	sprintf(gMsgStr, "shell: Starting service '%s'", args[1].c_str());
	my_cout(gMsgStr, kForceOut);
	
	afc_error_t ret = AMDeviceStartService(	sh->dev,
						CFStringCreateWithCString(	NULL,
						 				args[1].c_str(),
										kCFStringEncodingASCII),
						&sh->alt_service,
						NULL);
	
	if( !ret )
		my_cout("AMDeviceStartService: Service started");
	else
		my_cout("AMDeviceStartService: Service not found:", ret);
		
	return SHELL_CONTINUE;
	
}

int n_getfilesize(string *args, struct shell_state *sh)
{
	unsigned int size;
	string path;
	
	if( args[1] == "" )
	{
	
		// assume we want the size of the current remote dir
		path = sh->remote_path;
	} else if ( args[1].at(0) == '/' ) {
	
		// assume an abs path
		path = args[1];
	} else {
	
		// assume we want the size of something relative to remote_path
		path = sh->remote_path;
		processRelativePath(&path, &args[1]);
	}
	
	size = get_file_size(sh->conn, (char *) path.c_str());
	//cout << "getfilesize: " << size << " bytes" << endl;

	sprintf(gMsgStr, "//FILE_SIZE//%ld//<%s>//", (long) size, (char *) path.c_str());
	my_cout(gMsgStr, kForceOut);

	return SHELL_CONTINUE;
}

// CHK 2007/08/30
int n_status(string *args, struct shell_state *sh)
{
	int						ret, nSnapshots = 0;
	struct afc_directory	*dir;
	char					*dirent;
	
	// look for camera images

	if (dirExists(sh->conn, "/DCIM/100APPLE/"))
	{
		ret = AFCDirectoryOpen(sh->conn, "/DCIM/100APPLE/", &dir);  // open the dir for reading
		if (! ret)
		{
			while (1) 
			{
				AFCDirectoryRead(0, dir, &dirent);
				if (!dirent) break;
				
				if (strstr(dirent, ".THM"))
					++nSnapshots;
			}
		}
	}
	
	if (nSnapshots)
	{
		sprintf(gMsgStr, "//STATUS_HAVE_SNAPSHOTS//%d//", nSnapshots);
		my_cout(gMsgStr, kForceOut);
	}
	else
	{
		sprintf(gMsgStr, "//STATUS_NO_SNAPSHOTS//");
		my_cout(gMsgStr, kForceOut);
	}
	
	// check afc2:
	
	if (sh->shell_mode == SHELL_NORMAL)
		AFCConnectionClose(sh->conn);
	
	ret = AMDeviceStartService(sh->dev, CFStringCreateWithCString(NULL, "com.apple.afc2", kCFStringEncodingASCII), &(sh->afch), NULL);
	if (ret)
	{
		sprintf(gMsgStr, "//STATUS_AFC2_START_FAILED//%ld//", (long) ret);
		my_cout(gMsgStr, kForceOut);
		
		ret = AMDeviceStartService(sh->dev, CFStringCreateWithCString(NULL, "com.apple.afc", kCFStringEncodingASCII), &(sh->afch), NULL);
		if (ret)
		{
			sprintf(gMsgStr, "//STATUS_AFC_START_FAILED//%ld//", (long) ret);
			my_cout(gMsgStr, kForceOut);
		}
		else
		{
			ret = AFCConnectionOpen(sh->afch, 0, &(sh->conn));
			if (ret)
			{
				sprintf(gMsgStr, "//STATUS_AFC_OPEN_FAILED//%ld//", (long) ret);
				my_cout(gMsgStr, kForceOut);
			}
		}
		
		goto finish;
	}
	
	ret = AFCConnectionOpen(sh->afch, 0, &(sh->conn));
	if (ret)
	{
		sprintf(gMsgStr, "//STATUS_AFC2_OPEN_FAILED//%ld//", (long) ret);
		my_cout(gMsgStr, kForceOut);
		goto finish;
	}

	sprintf(gMsgStr, "//STATUS_AFC2_OK//");
	my_cout(gMsgStr, kForceOut);
	
	// phone statistics?
	// serial number?
	// phone number?

finish:
	return SHELL_TERMINATE;
}

int n_copyfile(string *args, struct shell_state *sh)
{
	string path1, path2;
	
	// make sure we're only talkin one file here
	if( dirExists(sh->conn, (char *)args[1].c_str() ) )
	{
		//cout << "//ERROR// " << "copyfile: Cannot retrieve '" << args[1]
		//	<< "' as it appears to be a directory." << endl;
		
		sprintf(gMsgStr, "//ERROR// copyfile: Cannot retrieve '%s' as it appears to be a directory.", args[1].c_str());
		my_cout(gMsgStr, kForceOut);
		return SHELL_CONTINUE;
	}
	
	if( args[1] == "" )
	{
		// assume we want current remote dir
		path1 = sh->remote_path;
		args[1] = path1;
	} else if ( args[1].at(0) == '/' ) {
		// assume an abs path
		path1 = args[1];
	} else {
		// assume we want something relative to remote_path
		path1 = sh->remote_path;
		processRelativePath(&path1, &args[1]);
		args[1] = path1;
	}

	if( args[2] == "" )
	{
		// assume we want current remote dir
		path2 = sh->remote_path;
		args[2] = path2;
	} else if ( args[2].at(0) == '/' ) {
		// assume an abs path
		path2 = args[2];
	} else {
		// assume we want something relative to remote_path
		path2 = sh->remote_path;
		processRelativePath(&path2, &args[2]);
		args[2] = path2;
	}
	
	copy_file(sh, (char *)args[1].c_str(), (char *)args[2].c_str());

	return SHELL_CONTINUE;
}

int n_backupfile(string *args, struct shell_state *sh)
{
	string path1, path2;
	
	// make sure we're only talkin one file here
	if( dirExists(sh->conn, (char *)args[1].c_str() ) )
	{
		//cout << "//ERROR// " << "backupfile: Cannot retrieve '" << args[1]
		//	<< "' as it appears to be a directory." << endl;
		
		sprintf(gMsgStr, "//ERROR// backupfile: Cannot retrieve '%s' as it appears to be a directory.", args[1].c_str());
		my_cout(gMsgStr, kForceOut);
		return SHELL_CONTINUE;
	}
	
	if( args[1] == "" )
	{
		// assume we want current remote dir
		path1 = sh->remote_path;
		args[1] = path1;
	} else if ( args[1].at(0) == '/' ) {
		// assume an abs path
		path1 = args[1];
	} else {
		// assume we want something relative to remote_path
		path1 = sh->remote_path;
		processRelativePath(&path1, &args[1]);
		args[1] = path1;
	}
	
	backup_file(sh, (char *)args[1].c_str());

	return SHELL_CONTINUE;
}

int n_unbackupfile(string *args, struct shell_state *sh)
{
	string path1, path2;
	
	// make sure we're only talkin one file here
	if( dirExists(sh->conn, (char *)args[1].c_str() ) )
	{
		//cout << "//ERROR// " << "unbackupfile: Cannot retrieve '" << args[1]
		//	<< "' as it appears to be a directory." << endl;
		
		sprintf(gMsgStr, "//ERROR// unbackupfile: Cannot retrieve '%s' as it appears to be a directory.", args[1].c_str());
		my_cout(gMsgStr, kForceOut);
		return SHELL_CONTINUE;
	}
	
	if( args[1] == "" )
	{
		// assume we want current remote dir
		path1 = sh->remote_path;
		args[1] = path1;
	} else if ( args[1].at(0) == '/' ) {
		// assume an abs path
		path1 = args[1];
	} else {
		// assume we want something relative to remote_path
		path1 = sh->remote_path;
		processRelativePath(&path1, &args[1]);
		args[1] = path1;
	}
	
	unbackup_file(sh, (char *)args[1].c_str());

	return SHELL_CONTINUE;
}

int n_restorefile(string *args, struct shell_state *sh)
{
	string path1, path2;
	
	// make sure we're only talkin one file here
	if( dirExists(sh->conn, (char *)args[1].c_str() ) )
	{
		//cout << "//ERROR// " << "restorefile: Cannot retrieve '" << args[1]
		//	<< "' as it appears to be a directory." << endl;
		
		sprintf(gMsgStr, "//ERROR// restorefile: Cannot retrieve '%s' as it appears to be a directory.", args[1].c_str());
		my_cout(gMsgStr, kForceOut);
		return SHELL_CONTINUE;
	}
	
	if( args[1] == "" )
	{
		// assume we want current remote dir
		path1 = sh->remote_path;
		args[1] = path1;
	} else if ( args[1].at(0) == '/' ) {
		// assume an abs path
		path1 = args[1];
	} else {
		// assume we want something relative to remote_path
		path1 = sh->remote_path;
		processRelativePath(&path1, &args[1]);
		args[1] = path1;
	}
	
	restore_file(sh, (char *)args[1].c_str());

	return SHELL_CONTINUE;
}

int n_getfile(string *args, struct shell_state *sh)
{
	string path;
	
	// make sure we're only talkin one file here
	if( dirExists(sh->conn, (char *)args[1].c_str() ) )
	{
		//cout << "//ERROR// " << "getfile: Cannot retrieve '" << args[1]
		//	<< "' as it appears to be a directory." << endl;
		
		sprintf(gMsgStr, "//ERROR// getfile: Cannot retrieve '%s' as it appears to be a directory.", args[1].c_str());
		my_cout(gMsgStr, kForceOut);
		return SHELL_CONTINUE;
	}
	
	if( args[1] == "" )
	{
	
		// assume we want current remote dir
		path = sh->remote_path;
	} else if ( args[1].at(0) == '/' ) {
	
		// assume an abs path
		path = args[1];
	} else {
	
		// assume we want something relative to remote_path
		path = sh->remote_path;
		processRelativePath(&path, &args[1]);
		args[1] = path;
	}
	
	// 2nd arg is optional
	if( args[2] == "" )
	{
		// if nothing is specified, simply copy to the current local directory
		args[2] = sh->local_path;
	
		// CHK 2007/08/10 (fixed bug in index to rfind)
			
		string temp = args[1].substr( args[1].rfind("/", args[1].length() - 1) + 1, args[1].length() - 1 );
		processRelativePath(&args[2], &temp);
	
	} else if (args[2].at(0) == '/') {
		
		// assume we want something relative to local_path
		string temp = sh->local_path;
		processRelativePath(&temp, &args[2]);
		args[2] = temp;
	}
	
//	cout << "remote: " << args[1] << endl;
//	cout << "local: " << args[2] << endl;
	
	get_file(sh, (char *)args[2].c_str(), (char *)args[1].c_str());
	
	return SHELL_CONTINUE;
}

int n_putfile(string *args, struct shell_state *sh)
{
	string path;
	
	if( args[1] == "" )
	{
		//cout << "//ERROR// " << "putfile: please specify a file to send" << endl;

		sprintf(gMsgStr, "//ERROR// putfile: please specify a file to send");
		my_cout(gMsgStr, kForceOut);
		return SHELL_CONTINUE;
		
	} else if ( args[1].at(0) == '/' ) {
	
		// assume an abs path
		path = args[1];
	} else {
	
		// assume we want something relative to local_path
		path = sh->local_path;
		processRelativePath(&path, &args[1]);
		args[1] = path;
	}
	
	// 2nd arg is optional
	if( args[2] == "" )
	{
		string temp;
		
		temp = args[1].substr( args[1].rfind("/",args[1].length() -1 ) + 1, (args[1].length() - 1) );
		args[2] = sh->remote_path;
		
		processRelativePath(&args[2], &temp);
	
	} else if (args[2].at(0) != '/') {

		// assume we want something relative to local_path
		string temp = sh->remote_path;
		processRelativePath(&temp, &args[2]);
		args[2] = temp;

	}

#if 1 // do this always ???
	backup_file1(sh, (char *) args[2].c_str());
#endif
	
	if (args[1] != "-backup") // if it *is* backup, then that's all we want...
	{
		put_file(sh, (char *)args[1].c_str(), (char *) args[2].c_str() );
	}
	
	return SHELL_CONTINUE;
}

int n_fileinfo(string *args, struct shell_state *sh)
{
	string path;
	char *key, *val;
	struct afc_dictionary *info;
	afc_error_t ret;
	
	if( args[1] == "" )
	{
		//cout << "//ERROR// " << "fileinfo: please provide a file to get info on." << endl;

		sprintf(gMsgStr, "//ERROR// fileinfo: please provide a file to get info on.");
		my_cout(gMsgStr, kForceOut);
		return SHELL_CONTINUE;
	} else if (args[1].at(0) != '/') {
		path = sh->remote_path;
		processRelativePath(&path, &args[1]);
	} else {
		path = sh->remote_path;
	}
	
	ret = AFCFileInfoOpen(sh->conn, (char *)path.c_str(), &info);
	if(ret)
	{
		//cout << "//ERROR// " << "AFCFileInfo: cannot open file '" << path << "': " << ret << endl;

		sprintf(gMsgStr, "//ERROR// AFCFileInfo: cannot open file '%s': %d", path.c_str(), ret);
		my_cout(gMsgStr, kForceOut);
		return SHELL_CONTINUE;
	}
	
	AFCKeyValueRead(info, &key, &val);
	while (key) {
		
		if (val)
		{
			//cout << key << ": " << val << endl;
			sprintf(gMsgStr, "%s: %s", key, val);
			my_cout(gMsgStr, kForceOut);
		}
		else
		{
			//cout << key << ": NULL" << endl;
			sprintf(gMsgStr, "%s: NULL", key);
			my_cout(gMsgStr, kForceOut);
		}
		
		AFCKeyValueRead(info, &key, &val);
	}
	AFCKeyValueClose(info);
	
	return SHELL_CONTINUE;
}

int n_exit(string *args, struct shell_state *sh)
{
	return SHELL_TERMINATE;
}

void put_file(struct shell_state *sh, char *local_path, char *remote_path)
{
	FILE *f;
	struct stat sb;
	unsigned char *buf;
	unsigned int size;
	afc_file_ref ref;
	std::string remotePathStr;
	afc_error_t ret;
	
	sprintf(gMsgStr, "putfile: Opening local file '%s'", local_path);
	my_cout(gMsgStr);
	
	f = fopen(local_path, "r");
	if (!f) {
		sprintf(gMsgStr, "putfile: Failed to open local file '%s'", local_path);
		my_cout(gMsgStr, kForceOutErr);
		return;
	}

	// If remote_path isn't set, use cp semantics
	if(remote_path == NULL)
	{
		// putfile foo/bar
		// should create a file named bar in the current dir
		string local = local_path;
		if(local[local.length() - 1] == '/')
		{
			my_cout("putfile: Writing directories is currently unsupported", kForceOutErr);
			return;
		}

		string::size_type lastSlash = local.find_last_of("/");
		if(lastSlash == string::npos)
		{
			// simple filename in local_path
			remotePathStr = local_path;
		} else {
			remotePathStr = local.substr(lastSlash + 1);
		}
	
	} else {
		remotePathStr = remote_path;
	}
	
	remote_path = (char *) remotePathStr.c_str();

	size = fstat(fileno(f), &sb);
	buf = (unsigned char *)malloc(sb.st_size);
	
	if(!buf)
	{
		my_cout("putfile: Could not allocate buffer. Aborting.", kForceOutErr);
		return;
	}
	
	fread(buf, sb.st_size, 1, f);
	fclose(f);
	
	sprintf(gMsgStr, "AFCFileRefOpen: opening remote file '%s'", remote_path);
	my_cout(gMsgStr);

	ret = AFCFileRefOpen(sh->conn, remote_path, 3, &ref);
	if (ret) 
	{
		sprintf(gMsgStr, "putfile: Failed to open remote file '%s'", remote_path);
		my_cout(gMsgStr, ret);
		return;
	}
	
	my_cout("AFCFileRefWrite: Writing to remote file");
	ret = AFCFileRefWrite(sh->conn, ref, buf, sb.st_size);
	
	if (ret) 
	{
		sprintf(gMsgStr, "putfile: Failed to write to remote file '%s'", remote_path);
		my_cout(gMsgStr, ret);
		return;
	}
	
	my_cout("AFCFileRefClose: Closing remote file.");
	ret = AFCFileRefClose(sh->conn, ref);
	
	if (ret) 
	{
		sprintf(gMsgStr, "putfile: Failed to close remote file '%s'", remote_path);
		my_cout(gMsgStr, ret);
		return;
	}

	free(buf);
	
	bSE_pushManifestItem(sh, kbSE_manifest_updated, local_path, remote_path);
	
	sprintf(gMsgStr, "//UPDATED//<%s>//<%s>//", local_path, remote_path);
	my_cout(gMsgStr, kForceOut);

	my_cout("putfile: Transfer successful");
}

// CHK 2007/08/23
void backup_file1(struct shell_state *sh, char *remote_path)
{
	unsigned char		*buf;
	std::string			remotePathStr;
	afc_error_t			ret;
	char				backupPath[1024];
	unsigned int		remote_file_size, backupFileSize;
	afc_file_ref		ref2, ref3;
	std::string			remFileName;
	
	// before overwriting a file on the phone, back it up to /_Backup_<remote_path>
	// if it doesn't already exist - i.e. preserve original files!

	if (remote_path == NULL)
		return;

	remotePathStr = remote_path;
	remote_path = (char *) remotePathStr.c_str();

	//---
	
	sprintf(backupPath, "/_Backup_%s", remote_path);
	
	// check if the file is already backed up
	backupFileSize = get_file_size(sh->conn, backupPath);
	if (! backupFileSize)
	{
		// check if the file to be 'put' exists locally
		ret = AFCFileRefOpen(sh->conn, remote_path, 1, &ref2);
		if (! ret) // it exists
		{
			remote_file_size = get_file_size(sh->conn, remote_path); 
			buf = (unsigned char *) malloc(remote_file_size);
			if (buf)
			{
				ret = AFCFileRefRead(sh->conn, ref2, buf, &remote_file_size);
				if (! ret)
				{
					AFCFileRefClose(sh->conn, ref2);
					
					ret = AFCFileRefOpen(sh->conn, backupPath, 3, &ref3);
					if (! ret)
					{
						AFCFileRefWrite(sh->conn, ref3, buf, remote_file_size);
						AFCFileRefClose(sh->conn, ref3);
						
						bSE_pushManifestItem(sh, kbSE_manifest_backedUp, remote_path, backupPath);
						
						sprintf(gMsgStr, "//BACKED UP//<%s>//", backupPath);
						my_cout(gMsgStr, kForceOut);
					}
				}
				
				free(buf);
			}
			
			AFCFileRefClose(sh->conn, ref2);
		}
	}
}

// CHK 2007/08/26
void copy_file(struct shell_state *sh, char *remote_path1, char *remote_path2)
{
	unsigned char		*buf;
	afc_error_t			ret;
	unsigned int		remote_file_size, dFileSize;
	afc_file_ref		ref2 = 0, ref3 = 0;
	
	// copy 'remote_path1' to 'remote_path2', but only if 'remote_path2' doesn't exist!
	
	if (remote_path1 == NULL)
		return;
	if (remote_path2 == NULL)
		return;

	// check if the file is already there
	dFileSize = get_file_size(sh->conn, remote_path2);
	if (! dFileSize)
	{
		// check if the file to be copied exists locally
		ret = AFCFileRefOpen(sh->conn, remote_path1, 1, &ref2);
		if (! ret) // it exists
		{
			remote_file_size = get_file_size(sh->conn, remote_path1); 
			buf = (unsigned char *) malloc(remote_file_size);
			if (buf)
			{
				ret = AFCFileRefRead(sh->conn, ref2, buf, &remote_file_size);
				if (! ret)
				{
					AFCFileRefClose(sh->conn, ref2);
					ref2 = 0;
					
					ret = AFCFileRefOpen(sh->conn, remote_path2, 3, &ref3);
					if (! ret)
					{
						AFCFileRefWrite(sh->conn, ref3, buf, remote_file_size);
						AFCFileRefClose(sh->conn, ref3);
						
						bSE_pushManifestItem(sh, kbSE_manifest_copied, remote_path1, remote_path2);
						
						sprintf(gMsgStr, "//COPIED//<%s>//<%s>//", remote_path1, remote_path2);
						my_cout(gMsgStr, kForceOut);
					}
				}
				
				free(buf);
			}
			
			if (ref2)
				AFCFileRefClose(sh->conn, ref2);
		}
	}
}

// CHK 2007/08/31
void unbackup_file(struct shell_state *sh, char *remote_path1)
{
	afc_error_t			ret;
	unsigned int		dFileSize;
	string::size_type	lastSlash;
	string				name, path, sremote_path1;
	char				remote_path2[1024];
	
	// restore file xxx at 'remote_path1' from _org_xxx
	// then delete _org_xxx
	
	if (remote_path1 == NULL)
		return;
	
	sremote_path1 = remote_path1;
	lastSlash = sremote_path1.find_last_of("/");
	name = sremote_path1.substr(lastSlash + 1);
	path = sremote_path1.substr(0, lastSlash + 1);
	sprintf(remote_path2, "%s_org_%s", path.c_str(), name.c_str());

	// check if the file is there
	dFileSize = get_file_size(sh->conn, remote_path2);
	if (dFileSize)
	{
		restore_file(sh, remote_path1);
		
		// now delete it
		ret = AFCRemovePath(sh->conn, remote_path2);
		if (! ret)
		{
			bSE_pushManifestItem(sh, kbSE_manifest_deleted, remote_path2, nil);
			
			sprintf(gMsgStr, "//DELETED//<%s>//", remote_path2);
			my_cout(gMsgStr, kForceOut);
		}
	}
}

// CHK 2007/08/30
void backup_file(struct shell_state *sh, char *remote_path1)
{
	unsigned char		*buf;
	afc_error_t			ret;
	unsigned int		remote_file_size, dFileSize;
	afc_file_ref		ref2 = 0, ref3 = 0;
	string::size_type	lastSlash;
	string				name, path, sremote_path1;
	char				remote_path2[1024];
	
	// copy file xxx at 'remote_path1' to _org_xxx, but only if it doesn't exist!
	
	if (remote_path1 == NULL)
		return;
	
	sremote_path1 = remote_path1;
	lastSlash = sremote_path1.find_last_of("/");
	name = sremote_path1.substr(lastSlash + 1);
	path = sremote_path1.substr(0, lastSlash + 1);
	sprintf(remote_path2, "%s_org_%s", path.c_str(), name.c_str());

	// check if the file is already there
	dFileSize = get_file_size(sh->conn, remote_path2);
	if (! dFileSize)
	{
		// check if the file to be copied exists locally
		ret = AFCFileRefOpen(sh->conn, remote_path1, 1, &ref2);
		if (! ret) // it exists
		{
			remote_file_size = get_file_size(sh->conn, remote_path1); 
			buf = (unsigned char *) malloc(remote_file_size);
			if (buf)
			{
				ret = AFCFileRefRead(sh->conn, ref2, buf, &remote_file_size);
				if (! ret)
				{
					AFCFileRefClose(sh->conn, ref2);
					ref2 = 0;
					
					ret = AFCFileRefOpen(sh->conn, remote_path2, 3, &ref3);
					if (! ret)
					{
						AFCFileRefWrite(sh->conn, ref3, buf, remote_file_size);
						AFCFileRefClose(sh->conn, ref3);
						
						bSE_pushManifestItem(sh, kbSE_manifest_backedUp, remote_path1, remote_path2);
						
						sprintf(gMsgStr, "//BACKED UP//<%s>//<%s>//", remote_path1, remote_path2);
						my_cout(gMsgStr, kForceOut);
					}
				}
				
				free(buf);
			}
			
			if (ref2)
				AFCFileRefClose(sh->conn, ref2);
		}
	}
}

// CHK 2007/08/30
void restore_file(struct shell_state *sh, char *remote_path1)
{
	unsigned char		*buf;
	afc_error_t			ret;
	unsigned int		remote_file_size;
	afc_file_ref		ref2 = 0, ref3 = 0;
	string::size_type	lastSlash;
	string				name, path, sremote_path1;
	char				remote_path2[1024];
	
	// restore file xxx at 'remote_path1' from _org_xxx
	
	if (remote_path1 == NULL)
		return;

	sremote_path1 = remote_path1;
	lastSlash = sremote_path1.find_last_of("/");
	name = sremote_path1.substr(lastSlash + 1);
	path = sremote_path1.substr(0, lastSlash + 1);
	sprintf(remote_path2, "%s_org_%s", path.c_str(), name.c_str());

	ret = AFCFileRefOpen(sh->conn, remote_path2, 1, &ref2);
	if (! ret) // it exists
	{
		remote_file_size = get_file_size(sh->conn, remote_path2); 
		buf = (unsigned char *) malloc(remote_file_size);
		if (buf)
		{
			ret = AFCFileRefRead(sh->conn, ref2, buf, &remote_file_size);
			if (! ret)
			{
				AFCFileRefClose(sh->conn, ref2);
				ref2 = 0;
				
				ret = AFCFileRefOpen(sh->conn, remote_path1, 3, &ref3);
				if (! ret)
				{
					AFCFileRefWrite(sh->conn, ref3, buf, remote_file_size);
					AFCFileRefClose(sh->conn, ref3);
					
					bSE_pushManifestItem(sh, kbSE_manifest_restored, remote_path1, remote_path2);
					
					sprintf(gMsgStr, "//RESTORED//<%s>//<%s>//", remote_path1, remote_path2);
					my_cout(gMsgStr, kForceOut);
				}
			}
			
			free(buf);
		}
		
		if (ref2)
			AFCFileRefClose(sh->conn, ref2);
	}
}

unsigned int get_file_size(struct afc_connection *conn, char *path)
{
	char *key, *val;
	unsigned int size;
	struct afc_dictionary *info;

	if (AFCFileInfoOpen(conn, path, &info))
	{
		sprintf(gMsgStr, "AFCFileInfoOpen: could not open file '%s'", path);
		my_cout(gMsgStr);
		return 0;
	}
	
	while (1) {
		AFCKeyValueRead(info, &key, &val);
		if (!key || !val)
			break;

		if (!strcmp(key, "st_size")) {
			sscanf(val, "%u", &size);
			AFCKeyValueClose(info);
			return size;
		}
	}

	AFCKeyValueClose(info);

	return 0;
}

void get_file(struct shell_state *sh, char *local_path, char *remote_path)
{
	FILE			*f = NULL;
	unsigned char	*buf = NULL;
	unsigned int	remote_file_size;
	afc_file_ref	ref = 0;
	afc_error_t		ret;
	
	// here we assume we are being passed absolute paths 
	// in both remote and local paths
	
	sprintf(gMsgStr, "AFCFileRefOpen: opening remote path '%s'", remote_path);
	my_cout(gMsgStr);

	ret = AFCFileRefOpen(sh->conn, remote_path, 1, &ref);
	if (ret)
	{
		sprintf(gMsgStr, "getfile: Failed to open remote file '%s'", remote_path);
		my_cout(gMsgStr, ret);
		goto finish;
	}
	
	remote_file_size = get_file_size(sh->conn, remote_path); 

	if ((remote_file_size > 8*1024*1024)
//		&&(0)
	)
	{
		unsigned int	copiedBytes, chunk;
		unsigned int 	maxChunk = 8*1024*1024;
		
		sprintf(gMsgStr, "AFCFileRefRead: copying %d bytes in chunks", remote_file_size);
		my_cout(gMsgStr, kForceOut);

		f = fopen(local_path, "w");
		if (f == NULL) {
			sprintf(gMsgStr, "getfile: Failed to open local file: '%s'", local_path);
			my_cout(gMsgStr, kForceOutErr);
			goto finish;
		}

		buf = (unsigned char *)malloc(maxChunk);
		if (!buf)
		{
			sprintf(gMsgStr, "getfile: could not allocate buffer. Aborting. %d '%s'", maxChunk, remote_path);
			my_cout(gMsgStr, kForceOutErr);
			goto finish;
		}

		copiedBytes = 0L;
		while(copiedBytes < remote_file_size)
		{
			chunk = maxChunk;

			if (copiedBytes+chunk >= remote_file_size)
				chunk = remote_file_size-copiedBytes;

			ret = AFCFileRefRead(sh->conn, ref, buf, &chunk);
			if (ret) 
			{
				sprintf(gMsgStr, "getfile: Failed to read from remote file: '%s'", remote_path);
				my_cout(gMsgStr, ret);
				goto finish;
			}
			
			fwrite(buf, chunk, 1, f);
			
			copiedBytes += chunk;
		}

		ret = AFCFileRefClose(sh->conn, ref);
		ref = 0;
		if (ret) 
		{
			sprintf(gMsgStr, "getfile: Failed to close remote file: '%s'", remote_path);
			my_cout(gMsgStr, ret);
			goto finish;
		}
	}
	else
	{
		sprintf(gMsgStr, "AFCFileRefRead: reading %d bytes into buffer", remote_file_size);
		my_cout(gMsgStr);

		buf = (unsigned char *)malloc(remote_file_size);
		if (!buf)
		{
			sprintf(gMsgStr, "getfile: could not allocate buffer. Aborting. %d '%s'", remote_file_size, remote_path);
			my_cout(gMsgStr, kForceOutErr);
			goto finish;
		}
		
		ret = AFCFileRefRead(sh->conn, ref, buf, &remote_file_size);
		if (ret) 
		{
			sprintf(gMsgStr, "getfile: Failed to read from remote file: '%s'", remote_path);
			my_cout(gMsgStr, ret);
			goto finish;
		}
	
		ret = AFCFileRefClose(sh->conn, ref);
		ref = 0;
		if (ret) 
		{
			sprintf(gMsgStr, "getfile: Failed to close remote file: '%s'", remote_path);
			my_cout(gMsgStr, ret);
			goto finish;
		}

		f = fopen(local_path, "w");
		if (f == NULL) {
			sprintf(gMsgStr, "getfile: Failed to open local file: '%s'", local_path);
			my_cout(gMsgStr, kForceOutErr);
			goto finish;
		}
		
		sprintf(gMsgStr, "getfile: Writing file to local path '%s'", local_path);
		my_cout(gMsgStr);
		
		fwrite(buf, remote_file_size, 1, f);
	}

	my_cout("getfile: Transfer successful.");

finish:
	
	if (ref)	AFCFileRefClose(sh->conn, ref);
	if (buf)	free(buf);
	if (f)		fclose(f);
	
	return;
}

/* -*- mode:c; indent-tabs-mode:nil; c-basic-offset:2; tab-width:2; */
